/***********************************************************************************
 *                                                                                 *
 * Voreen - The Volume Rendering Engine                                            *
 *                                                                                 *
 * Copyright (C) 2005-2020 University of Muenster, Germany,                        *
 * Department of Computer Science.                                                 *
 * For a list of authors please refer to the file "CREDITS.txt".                   *
 *                                                                                 *
 * This file is part of the Voreen software package. Voreen is free software:      *
 * you can redistribute it and/or modify it under the terms of the GNU General     *
 * Public License version 2 as published by the Free Software Foundation.          *
 *                                                                                 *
 * Voreen is distributed in the hope that it will be useful, but WITHOUT ANY       *
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR   *
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.      *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License in the file   *
 * "LICENSE.txt" along with this file. If not, see <http://www.gnu.org/licenses/>. *
 *                                                                                 *
 * For non-commercial academic use see the license exception specified in the file *
 * "LICENSE-academic.txt". To get information about commercial licensing please    *
 * contact the authors.                                                            *
 *                                                                                 *
 ***********************************************************************************/

#include "resamplefilterproperties.h"
#include "../volumefiltering/resamplefilter.h"

namespace voreen {

#define RESAMPLE_DIM_DEFAULT_VALUE tgt::ivec3(100000)

ResampleFilterProperties::ResampleFilterProperties()
    : dimensions_(getId("dimensions"), "Target Dimensions", RESAMPLE_DIM_DEFAULT_VALUE, tgt::ivec3(2), tgt::ivec3(100000))
{
    // Store default settings.
    storeInstance(DEFAULT_SETTINGS);

    // Add properties to list.
    addProperties();
}

std::string ResampleFilterProperties::getVolumeFilterName() const {
    return "Resample Filter";
}

void ResampleFilterProperties::adjustPropertiesToInput(const VolumeBase& input) {
    if(dimensions_.get() == RESAMPLE_DIM_DEFAULT_VALUE) {
        dimensions_.set(input.getDimensions());
    }
}

VolumeFilter* ResampleFilterProperties::getVolumeFilter(const VolumeBase& volume, int instanceId) const {
    if (instanceSettings_.find(instanceId) == instanceSettings_.end()) {
        return nullptr;
    }
    Settings settings = instanceSettings_.at(instanceId);
    return new ResampleFilter(
        settings.dimensions_,
        volume.getBaseType(), volume.getNumChannels()
    );
}
void ResampleFilterProperties::restoreInstance(int instanceId) {
    auto iter = instanceSettings_.find(instanceId);
    if (iter == instanceSettings_.end()) {
        instanceSettings_[instanceId] = instanceSettings_[DEFAULT_SETTINGS];
    }

    Settings settings = instanceSettings_[instanceId];
    dimensions_.set(settings.dimensions_);
}
void ResampleFilterProperties::storeInstance(int instanceId) {
    Settings& settings = instanceSettings_[instanceId];
    settings.dimensions_ = dimensions_.get();
}
void ResampleFilterProperties::removeInstance(int instanceId) {
    instanceSettings_.erase(instanceId);
}
void ResampleFilterProperties::addProperties() {
    properties_.push_back(&dimensions_);
}
void ResampleFilterProperties::serialize(Serializer& s) const {
    s.serialize(getId("instanceSettings"), instanceSettings_);
}
void ResampleFilterProperties::deserialize(Deserializer& s) {
    try {
        s.deserialize(getId("instanceSettings"), instanceSettings_);
    }
    catch (SerializationException&) {
        s.removeLastError();
        LERROR("You need to reconfigure " << getVolumeFilterName() << " instances of " << ( properties_[0]->getOwner() ? properties_[0]->getOwner()->getGuiName() : "VolumeFilterList"));
    }
}
std::vector<int> ResampleFilterProperties::getStoredInstances() const {
    std::vector<int> output;
    for(auto& kv : instanceSettings_) {
        if(kv.first != DEFAULT_SETTINGS) {
            output.push_back(kv.first);
        }
    }
    return output;
}

void ResampleFilterProperties::Settings::serialize(Serializer& s) const {
    s.serialize("dimensions", dimensions_);
}
void ResampleFilterProperties::Settings::deserialize(Deserializer& s) {
    s.deserialize("dimensions", dimensions_);
}
}
